==> viset-0.1.13.tar.gz: viset-0.1.13/viset/cache.py <==
"""Object cache for URLs with garbage collection"""

import os
from os import path
import hashlib
import urlparse
from viset.util import isarchive, isurl, isimg, ishdf5, isfile
import viset.download
import pylab
import string
import shutil

class Cache():
    _cacheroot = os.environ.get('VISYM_CACHE')
    if _cacheroot is None:
        _cacheroot = path.join(os.environ['HOME'],'.visym','cache')
    if not path.exists(_cacheroot):
        os.makedirs(_cacheroot)
    _maxsize = None
    _verbose = None
    _strategy = None
    _free_maxctr = 100
    _free_ctr = _free_maxctr
    _cachesize = None  # async result
    _prettyhash = True
    
    def __init__(self, cacheroot=_cacheroot, maxsize=10E9, verbose=False, strategy='lru'):
        if cacheroot is not None:
            self._cacheroot = cacheroot
        self._maxsize = maxsize
        self._verbose = verbose
        self._strategy = strategy
            
    def __len__(self):
        if self._cachesize is not None:
            return self._cachesize.get()
        else:
            return self.size()
        
    def __str__(self):
        return str('this is a string')


    def _url_fragment_options(self, url, sha1=None):
        default_opts = {'sha1':sha1, 'item':None, 'reader':None, 'id':None}
        url_scheme = urlparse.urlparse(url)[0]
        url_fragment = urlparse.urlparse(url)[5]   
        opt = default_opts
        fragment = urlparse.parse_qs(url_fragment)
        for key in fragment.keys():
            if type(fragment[key]) is list:
                opt[key] = fragment[key][0]
            else:
                opt[key] = fragment[key]                
        return opt
  
    def read(self, filename, reader=None, id=None):
        quietprint('[viset.cache]: Reading ' + reader + ' encoded image ' + str(id) + ' from file "' + filename + '"', self._verbose)
        if reader == 'universal':
            return pylab.imread(filename)
        elif reader == 'mnist':
            return viset.mnist.imread(filename, int(id))
        else:
            self.quietprint('[viset.cache]: reader type "' + str(reader) + '" unsupported', self._verbose)
            raise ValueError()
        
    def unpack(self, filename, sha1=None, item=None):
        """Extract cached archive to archive directory file and delete archive"""
        if isarchive(filename):
            (unpackdir, ext) = viset.util.splitextension(filename)
            if not path.exists(unpackdir):
                os.makedirs(unpackdir)
            viset.download.extract(filename, unpackdir, sha1=sha1, verbose=self._verbose)            
            os.remove(filename)
            if item is not None:
                filename = path.join(unpackdir, item)
            else:
                filename = unpackdir    
        return filename
        
    def load(self, filename, reader=None, id=None):
        """Load a filename to an object"""
        if isimg(filename):
            obj = pylab.imread(filename)
        elif reader is not None:
            obj = self.read(filename, reader=reader, id=id)
        elif path.isdir(viset.util.splitextension(filename)[0]):
            obj = viset.util.splitextension(filename)[0]  # archive directory in cache
        elif isarchive(filename):
            quietprint('[viset.cache]: archive file "' + filename + '" not unpacked to "' + (viset.util.splitextension(filename)[0]), self._verbose)
            raise IOError()
        elif ishdf5(filename):
            obj = filename
        elif isfile(filename):
            obj = filename
        else:
            quietprint('[viset.cache]: could not load filename "' + filename + '"', self._verbose)
            raise IOError()
        return obj
        
        
    def download(self, url):        
        """Download element according to URL scheme"""
        self.free()  # garbage collection time?        
        filename = self.cachefile(urlparse.urldefrag(url)[0])
        url_scheme = urlparse.urlparse(url)[0]
        if url_scheme == 'http':
            viset.download.download(url, filename, verbose=self._verbose)                       
        elif url_scheme == 'file':
            pass
        elif url_scheme == 'viset':
            raise IOError('FIXME: support for viset database queries')                            
        elif url_scheme == 'hdfs':
            raise IOError('FIXME: support for hadoop distributed file system')                
        else:
            raise IOError('FIXME: support for URL scheme ' + url_scheme)            
        return filename        

    def delete(self):
        """Delete entire cache"""
        quietprint('[viset.cache]: Deleting all cached data in "' + self._cacheroot + '"', self._verbose)
        shutil.rmtree(self._cacheroot)
        os.makedirs(self._cacheroot)        
        
    def fetch(self, url, sha1=None):
        """Fetch and element and store in cache"""
        if not isurl(url):
            return None
        urlopts = self._url_fragment_options(url, sha1)
        if not self.iscached(url):
            quietprint('[viset.cache][MISS]: downloading \'' + str(url[0:63] + '...') + '\'', self._verbose)                                        
            filename = self.download(url)
            if isarchive(filename):
                quietprint('[viset.cache]: extracting archive \'' + filename + '\'', self._verbose)
                if urlopts['item'] is not None:
                    filename = self.unpack(filename, sha1=urlopts['sha1'], item=urlopts['item'])
                else:
                    filename = self.unpack(filename, sha1=urlopts['sha1'], item=None)                        
        else:
            filename = self.cachefile(url)            
            quietprint('[viset.cache][HIT]: retrieving \'' + filename + '\'', self._verbose)                                     
        quietprint('[viset.cache]: loading \'' + filename + '\'', self._verbose)                                         
        return self.load(filename, reader=urlopts['reader'], id=urlopts['id'])
    
    def get(self, url, sha1=None):
        """Get an element from cache"""        
        try:
            return self.fetch(url, sha1=sha1)
        except ValueError:
            raise
        except:
            quietprint('[viset.cache]: cache fetch error - discarding and refetching', self._verbose)                                                     
            self.discard(url)
            return self.fetch(url, sha1=sha1)
            
    def discard(self, url):
        """Remove element from cache"""
        if path.isfile(self.cachefile(url)):
            os.remove(self.cachefile(url))
        if path.isdir(self.cachedir(url)):
            shutil.rmtree(self.cachedir(url))
        
    def size(self, source=_cacheroot):
        """Recursively compute the size of a directory: http://snipplr.com/view/47686/"""
        total_size = os.path.getsize(source)
        for item in os.listdir(source):
            itempath = os.path.join(source, item)
            if os.path.isfile(itempath):
                total_size += os.path.getsize(itempath)
            elif os.path.isdir(itempath):
                total_size += self.size(itempath)
        return total_size

    def hash(self, url, prettyhash=_prettyhash):
        """Compute a SHA1 file hash to uniquely name a cached file"""
        p = urlparse.urlsplit(url)
        urlquery = urlparse.urlunsplit([p[0],p[1],p[2],p[3],None])
        urlpath = urlparse.urlunsplit([p[0],p[1],p[2],None,None])        
        (filename, ext) = viset.util.splitextension(urlpath)
        urlopt = self._url_fragment_options(url)
        urlhash = hashlib.sha1(urlquery).hexdigest()
        if prettyhash:    
            return path.basename(filename) + '_' + urlhash[0:7]
        else:
            return urlhash 

        
    def free(self):
        """Garbage collection"""
        if self._free_ctr == 0:
            if self._cachesize is not None:
                if self._cachesize.get() > self._maxsize:
                    print 'WARNING: cachesize is larger than maximum.  Clean resources!'
            quietprint('[viset.cache.free]: spawning cache garbage collection process', self._verbose)
            self._cachesize = Pool(1).apply_async(self.size(), self._cacheroot)
            self._free_ctr = self._free_maxctr
        self._free_ctr -= 1
    
    def cachefile(self, url):
        """The file name of the element in the cache"""
        p = urlparse.urlsplit(url)
        urlquery = urlparse.urlunsplit([p[0],p[1],p[2],p[3],None])        
        urlpath = urlparse.urlunsplit([p[0],p[1],p[2],None,None])
        urlopt = self._url_fragment_options(url)
        urlhash = self.hash(url)
        if urlopt['item'] is not None:
            return self.abspath(path.join(urlhash, urlopt['item']))
        else:
            (filename, ext) = viset.util.splitextension(path.basename(urlpath))
            return self.abspath(urlhash + ext)

    def cachedir(self, archive_url):
        """The unpacked directory name of archive files in the cache"""
        p = urlparse.urlsplit(archive_url)
        urlpath = urlparse.urlunsplit([p[0],p[1],p[2],p[3],None])
        urlopt = self._url_fragment_options(archive_url)
        urlhash = self.hash(urlpath)
        return self.abspath(urlhash)
        
    def abspath(self, filename):
        """The absolute file path for elements in the cache"""
        return path.join(self._cacheroot, filename)
        
    def iscached(self, url):
        """Return true if an element is in the cache"""
        return path.isfile(self.cachefile(url)) or path.isdir(self.cachedir(url))
            
            
def quietprint(mystr, is_verbose):
    if is_verbose:
        print mystr
            
            
##     convert from viset:// to file:// or http://
##     viset:path_to_resource
##     viset://domain.com/

##     cannot access http://path.to.h5?query since a row may contain a viset.cache which cannot be handled 
##     viset://dbname.h5?view=''&group=''&table=''&row=''&field=''
##     viset://dbname.h5?path='/root/annotation/categorization/0/url'
##     URI: viset.cache:/path/to/resource => URL FILE://{VISET_CACHE}/path/to/resource

##     viset.cache.mnist:path/to/mnistbinary.gz?id=0

##     can still include relative uri in the viset.  relative uris are relative to base uri FILE://{VISET_CACHE}/path

##     viset.http://domain.com/path.h5
##     viset.hdfs:path.h5?q='query'

##     viset:/absolute/path.h5?im=0   
##     viset:path.h5?im=0    # relative url, get in cache
##     viset://domain.com/absolute/path.h5?view='image'&id=0        # query server for just row
##     viset.http://domain.com/absolute/path.h5?view='image'&id=0    # download entire viset via http
##     viset.hdfs://domain.com/absolute/path.h5?view='categorization'&id=0        # access viset via hdfs

##     viset.mnist:/path.gz#frag  pass to viset mnist handler for reading 
    
##     do we just cache function calls to viset.types.Image object?  no, want to make sure that separate calls to Image from same package use cached data
##     use queries and fragments
##     querystrings are passed to the server to return a document based on the query.  this query file is returned from the cache
##     fragments are handled by the cache itself.  
##     who unzips? who loads?  file cache vs. object cache?
    
##     fragments are passed to
    
##     image and annotations are handled by viset uris.  don't bother importing on the client and pickling, send viset uri and require the client to resolve to locations and unpickle
    
##     foo://username:password@example.com:8042/over/there/index.dtb?type=animal&name=narwhal#nose

##     if we ahve a local copy of a viset, how do we share with clients?  need to copy it somehow, or pass a reference.
## how can we still pass rows?
## must put the file in a place accessible by clients
## maybe expose viset://master.ip/path.h5
## no, pass package dependencies along with the rows so we download only those packages necessary
## packages must be accessible via http (by construction of viset)

## just pass the package url to the clients
## viset.http://packageurl.tgz#fragment
## viset.dbname: relative url

## do not store separate packages, store urls with fragments.  the caching takes care of downloading and unpacking once we use it

## viset://package.tgz?reader=mnist&id=0   cache hashes url (without querystring) to filename for reuse between visets 
## viset defaults to http, but need viset scheme to handle querystring appropriately .  use viset.hdfs for nested schemes (inner fetches file, outer handles query and fragment)
## client does not know or care what database this came from.  it downloads and caches for everyone unless overridden for private cache

## caching on shared machine?  winner locks the file and os cannot open it for writing.  wait and retest if cached

## https://github.com/smontanaro/pylockfile
## http://docs.python.org/2/library/fcntl.html#fcntl.lockf
## http://www.evanfosmark.com/2009/01/cross-platform-file-locking-support-in-python/
## http://amix.dk/blog/post/19531

## viset image object has a standard viset: scheme for caching and importing and image

## for shared resources, can pass URL FILE://
## never see any relative urls stored in viset everything is a URL and caching is transparent

==> viset-0.1.14.tar.gz: viset-0.1.14/viset/cache.py <==
"""Object cache for URLs with garbage collection"""

import os
from os import path
import hashlib
import urlparse
from viset.util import isarchive, isurl, isimg, ishdf5, isfile
import viset.download
import pylab
import string
import shutil

class Cache():
    _cacheroot = os.environ.get('VISYM_CACHE')
    if _cacheroot is None:
        _cacheroot = path.join(os.environ['HOME'],'.visym','cache')
    if not path.exists(_cacheroot):
        os.makedirs(_cacheroot)
    _maxsize = None
    _verbose = None
    _strategy = None
    _free_maxctr = 100
    _free_ctr = _free_maxctr
    _cachesize = None  # async result
    _prettyhash = True
    
    def __init__(self, cacheroot=_cacheroot, maxsize=10E9, verbose=False, strategy='lru'):
        if cacheroot is not None:
            self._cacheroot = cacheroot
        self._maxsize = maxsize
        self._verbose = verbose
        self._strategy = strategy
            
    def __len__(self):
        if self._cachesize is not None:
            return self._cachesize.get()
        else:
            return self.size()
        
    def __str__(self):
        return str('this is a string')


    def _url_fragment_options(self, url, sha1=None):
        default_opts = {'sha1':sha1, 'item':None, 'reader':None, 'id':None}
        url_scheme = urlparse.urlparse(url)[0]
        url_fragment = urlparse.urlparse(url)[5]   
        opt = default_opts
        fragment = urlparse.parse_qs(url_fragment)
        for key in fragment.keys():
            if type(fragment[key]) is list:
                opt[key] = fragment[key][0]
            else:
                opt[key] = fragment[key]                
        return opt
  
    def read(self, filename, reader=None, id=None):
        quietprint('[viset.cache]: Reading ' + reader + ' encoded image ' + str(id) + ' from file "' + filename + '"', self._verbose)
        if reader == 'universal':
            return pylab.imread(filename)
        elif reader == 'mnist':
            return viset.mnist.imread(filename, int(id))
        else:
            self.quietprint('[viset.cache]: reader type "' + str(reader) + '" unsupported', self._verbose)
            raise NotImplementedError()
        
    def unpack(self, filename, sha1=None, item=None):
        """Extract cached archive to archive directory file and delete archive"""
        if isarchive(filename):
            (unpackdir, ext) = viset.util.splitextension(filename)
            if not path.exists(unpackdir):
                os.makedirs(unpackdir)
            viset.download.extract(filename, unpackdir, sha1=sha1, verbose=self._verbose)            
            os.remove(filename)
            if item is not None:
                filename = path.join(unpackdir, item)
            else:
                filename = unpackdir    
        return filename
        
    def load(self, filename, reader=None, id=None):
        """Load a filename to an object"""
        if isimg(filename):
            obj = pylab.imread(filename)
        elif reader is not None:
            obj = self.read(filename, reader=reader, id=id)
        elif path.isdir(viset.util.splitextension(filename)[0]):
            obj = viset.util.splitextension(filename)[0]  # archive directory in cache
        elif isarchive(filename):
            quietprint('[viset.cache]: archive file "' + filename + '" not unpacked to "' + (viset.util.splitextension(filename)[0]), self._verbose)
            raise IOError()
        elif ishdf5(filename):
            obj = filename
        elif isfile(filename):
            obj = filename
        else:
            quietprint('[viset.cache]: could not load filename "' + filename + '"', self._verbose)
            raise IOError()
        return obj
        
        
    def download(self, url):        
        """Download element according to URL scheme"""
        self.free()  # garbage collection time?        
        filename = self.cachefile(urlparse.urldefrag(url)[0])
        url_scheme = urlparse.urlparse(url)[0]
        if url_scheme == 'http':
            viset.download.download(url, filename, verbose=self._verbose)                       
        elif url_scheme == 'file':
            pass
        elif url_scheme == 'viset':
            raise NotImplementedError('FIXME: support for viset database queries')                            
        elif url_scheme == 'hdfs':
            raise NotImplementedError('FIXME: support for hadoop distributed file system')                
        else:
            raise NotImplementedError('FIXME: support for URL scheme ' + url_scheme)            
        return filename        

    def delete(self):
        """Delete entire cache"""
        quietprint('[viset.cache]: Deleting all cached data in "' + self._cacheroot + '"', self._verbose)
        shutil.rmtree(self._cacheroot)
        os.makedirs(self._cacheroot)        
        
    def fetch(self, url, sha1=None):
        """Fetch and element and store in cache"""
        if not isurl(url):
            return None
        urlopts = self._url_fragment_options(url, sha1)
        if not self.iscached(url):
            quietprint('[viset.cache][MISS]: downloading \'' + str(url[0:63] + '...') + '\'', self._verbose)                                        
            filename = self.download(url)
            if isarchive(filename):
                quietprint('[viset.cache]: extracting archive \'' + filename + '\'', self._verbose)
                if urlopts['item'] is not None:
                    filename = self.unpack(filename, sha1=urlopts['sha1'], item=urlopts['item'])
                else:
                    filename = self.unpack(filename, sha1=urlopts['sha1'], item=None)                        
        else:
            filename = self.cachefile(url)            
            quietprint('[viset.cache][HIT]: retrieving \'' + filename + '\'', self._verbose)                                     
        quietprint('[viset.cache]: loading \'' + filename + '\'', self._verbose)                                         
        return self.load(filename, reader=urlopts['reader'], id=urlopts['id'])
    
    def get(self, url, sha1=None):
        """Get an element from cache"""        
        try:
            return self.fetch(url, sha1=sha1)
        except KeyboardInterrupt:
            pass
        except NotImplementedError:
            raise
        except:
            quietprint('[viset.cache]: cache fetch error - discarding and refetching', self._verbose)                                                     
            self.discard(url)
            return self.fetch(url, sha1=sha1)
            
    def discard(self, url):
        """Remove element from cache"""
        if path.isfile(self.cachefile(url)):
            os.remove(self.cachefile(url))
        if path.isdir(self.cachedir(url)):
            shutil.rmtree(self.cachedir(url))
        
    def size(self, source=_cacheroot):
        """Recursively compute the size of a directory: http://snipplr.com/view/47686/"""
        total_size = os.path.getsize(source)
        for item in os.listdir(source):
            itempath = os.path.join(source, item)
            if os.path.isfile(itempath):
                total_size += os.path.getsize(itempath)
            elif os.path.isdir(itempath):
                total_size += self.size(itempath)
        return total_size

    def hash(self, url, prettyhash=_prettyhash):
        """Compute a SHA1 file hash to uniquely name a cached file"""
        p = urlparse.urlsplit(url)
        urlquery = urlparse.urlunsplit([p[0],p[1],p[2],p[3],None])
        urlpath = urlparse.urlunsplit([p[0],p[1],p[2],None,None])        
        (filename, ext) = viset.util.splitextension(urlpath)
        urlopt = self._url_fragment_options(url)
        urlhash = hashlib.sha1(urlquery).hexdigest()
        if prettyhash:    
            return path.basename(filename) + '_' + urlhash[0:7]
        else:
            return urlhash 

        
    def free(self):
        """Garbage collection"""
        if self._free_ctr == 0:
            if self._cachesize is not None:
                if self._cachesize.get() > self._maxsize:
                    print 'WARNING: cachesize is larger than maximum.  Clean resources!'
            quietprint('[viset.cache.free]: spawning cache garbage collection process', self._verbose)
            self._cachesize = Pool(1).apply_async(self.size(), self._cacheroot)
            self._free_ctr = self._free_maxctr
        self._free_ctr -= 1
    
    def cachefile(self, url):
        """The file name of the element in the cache"""
        p = urlparse.urlsplit(url)
        urlquery = urlparse.urlunsplit([p[0],p[1],p[2],p[3],None])        
        urlpath = urlparse.urlunsplit([p[0],p[1],p[2],None,None])
        urlopt = self._url_fragment_options(url)
        urlhash = self.hash(url)
        if urlopt['item'] is not None:
            return self.abspath(path.join(urlhash, urlopt['item']))
        else:
            (filename, ext) = viset.util.splitextension(path.basename(urlpath))
            return self.abspath(urlhash + ext)

    def cachedir(self, archive_url):
        """The unpacked directory name of archive files in the cache"""
        p = urlparse.urlsplit(archive_url)
        urlpath = urlparse.urlunsplit([p[0],p[1],p[2],p[3],None])
        urlopt = self._url_fragment_options(archive_url)
        urlhash = self.hash(urlpath)
        return self.abspath(urlhash)
        
    def abspath(self, filename):
        """The absolute file path for elements in the cache"""
        return path.join(self._cacheroot, filename)
        
    def iscached(self, url):
        """Return true if an element is in the cache"""
        return path.isfile(self.cachefile(url)) or path.isdir(self.cachedir(url))
            
            
def quietprint(mystr, is_verbose):
    if is_verbose:
        print mystr
            
            
##     convert from viset:// to file:// or http://
##     viset:path_to_resource
##     viset://domain.com/

##     cannot access http://path.to.h5?query since a row may contain a viset.cache which cannot be handled 
##     viset://dbname.h5?view=''&group=''&table=''&row=''&field=''
##     viset://dbname.h5?path='/root/annotation/categorization/0/url'
##     URI: viset.cache:/path/to/resource => URL FILE://{VISET_CACHE}/path/to/resource

##     viset.cache.mnist:path/to/mnistbinary.gz?id=0

##     can still include relative uri in the viset.  relative uris are relative to base uri FILE://{VISET_CACHE}/path

##     viset.http://domain.com/path.h5
##     viset.hdfs:path.h5?q='query'

##     viset:/absolute/path.h5?im=0   
##     viset:path.h5?im=0    # relative url, get in cache
##     viset://domain.com/absolute/path.h5?view='image'&id=0        # query server for just row
##     viset.http://domain.com/absolute/path.h5?view='image'&id=0    # download entire viset via http
##     viset.hdfs://domain.com/absolute/path.h5?view='categorization'&id=0        # access viset via hdfs

##     viset.mnist:/path.gz#frag  pass to viset mnist handler for reading 
    
##     do we just cache function calls to viset.types.Image object?  no, want to make sure that separate calls to Image from same package use cached data
##     use queries and fragments
##     querystrings are passed to the server to return a document based on the query.  this query file is returned from the cache
##     fragments are handled by the cache itself.  
##     who unzips? who loads?  file cache vs. object cache?
    
##     fragments are passed to
    
##     image and annotations are handled by viset uris.  don't bother importing on the client and pickling, send viset uri and require the client to resolve to locations and unpickle
    
##     foo://username:password@example.com:8042/over/there/index.dtb?type=animal&name=narwhal#nose

##     if we ahve a local copy of a viset, how do we share with clients?  need to copy it somehow, or pass a reference.
## how can we still pass rows?
## must put the file in a place accessible by clients
## maybe expose viset://master.ip/path.h5
## no, pass package dependencies along with the rows so we download only those packages necessary
## packages must be accessible via http (by construction of viset)

## just pass the package url to the clients
## viset.http://packageurl.tgz#fragment
## viset.dbname: relative url

## do not store separate packages, store urls with fragments.  the caching takes care of downloading and unpacking once we use it

## viset://package.tgz?reader=mnist&id=0   cache hashes url (without querystring) to filename for reuse between visets 
## viset defaults to http, but need viset scheme to handle querystring appropriately .  use viset.hdfs for nested schemes (inner fetches file, outer handles query and fragment)
## client does not know or care what database this came from.  it downloads and caches for everyone unless overridden for private cache

## caching on shared machine?  winner locks the file and os cannot open it for writing.  wait and retest if cached

## https://github.com/smontanaro/pylockfile
## http://docs.python.org/2/library/fcntl.html#fcntl.lockf
## http://www.evanfosmark.com/2009/01/cross-platform-file-locking-support-in-python/
## http://amix.dk/blog/post/19531

## viset image object has a standard viset: scheme for caching and importing and image

## for shared resources, can pass URL FILE://
## never see any relative urls stored in viset everything is a URL and caching is transparent
